@@ -24,30 +24,16 @@ __all__ = ["HASHFILE_NAME","IntegrityCheckError",
            "SignedImportManager","RSAKeyWithPSS"]
 
 
-#  Careful now, we can't just import things willy-nilly.  The following
-#  tries to grab the things we absolutely need without polluting the
-#  global namespace.
+#  Careful now, we can't just import things willy-nilly.  Make sure you
+#  only use builtin modules at the top level.
 
-def _signedimp_mod_available(modname):
-    """Check whether the named module is safely available.
-
-    To be safe it must be either built into the interpreter, or already
-    loaded in sys.modules.
-    """
-    if modname in sys.builtin_module_names:
-        return True
-    if sys.modules.get(modname) is not None:
-        return True
-    if imp.is_frozen(modname):
-        return True
-    return False
 
 #  The sys module is always builtin, since it's needed to do imports.
 import sys
 
 #  Get the "imp" module so we can simulate the standard import machinery.
 #  It must be builtin or this whole exercise is pointless.
-if not _signedimp_mod_available("imp"):
+if "imp" not in sys.builtin_module_names:
     err = "'imp' module is not safely available, integrity checks impossible"
     raise IntegrityCheckMissing(err)
 import imp
@@ -55,7 +41,7 @@ import imp
 
 #  Get the "marshal" module so we can treat code objects as bytes.
 #  It must be builtin or this whole exercise is pointless.
-if not _signedimp_mod_available("marshal"):
+if "marshal" not in sys.builtin_module_names:
     err = "'marshal' module is not safely available, " \
           "integrity checks impossible"
     raise IntegrityCheckMissing(err)
@@ -65,13 +51,10 @@ import marshal
 #  Get a minimal simulation of the "os" module.
 #  It must be builtin or this whole exercise is pointless.
 def _signedimp_make_os_module():
-    if _signedimp_mod_available("os") and _signedimp_mod_available("os.path"):
-        import os
-        return os
-    if _signedimp_mod_available("posix"):
+    if "posix" in sys.builtin_module_names:
         from posix import stat
         SEP = "/"
-    elif _signedimp_mod_available("nt"):
+    elif "nt" in sys.builtin_module_names:
         from nt import stat
         SEP = "\\"
     else:
@@ -83,7 +66,12 @@ def _signedimp_make_os_module():
             @staticmethod
             def join(*args):
                 """Local re-implementation of os.path.join."""
-                return SEP.join(args)
+                bits = []
+                for bit in args:
+                    while bit.endswith(SEP) and bit != SEP:
+                        bit = bit[:-1]
+                    bits.append(bit)
+                return SEP.join(bits)
             @staticmethod
             def exists(path):
                 """Local re-implementation of os.path.exists."""
@@ -92,6 +80,10 @@ def _signedimp_make_os_module():
                 except EnvironmentError:
                     return False
                 return True
+            @staticmethod
+            def dirname(p):
+                """Local re-implementation of os.path.dirname."""
+                return SEP.join(p.split(SEP)[:-1])
     return os
 os = _signedimp_make_os_module()
 
@@ -131,6 +123,8 @@ class _signedimp_util:
         if hashlib is not available safely.  It uses the pure-python MD5 code
         from the PyPy project, which is available under the Python License.
         """
+        #  The signedimp.tools.get_bootstrap_code() function will inline the
+        #  raw code from signedimp.cryptobase.md5
         if not _signedimp_util._md5type:
             from signedimp.cryptobase.md5 import MD5Type
             _signedimp_util._md5type = MD5Type
@@ -145,6 +139,8 @@ class _signedimp_util:
         if hashlib is not available safely.  It uses the pure-python SHA1 code
         from the PyPy project, which is available under the Python License.
         """
+        #  The signedimp.tools.get_bootstrap_code() function will inline the
+        #  raw code from signedimp.cryptobase.sha1
         if not _signedimp_util._sha1type:
             from signedimp.cryptobase.sha1 import sha1
             _signedimp_util._sha1type = sha1
@@ -158,6 +154,8 @@ class _signedimp_util:
         This is horribly slow and probably broken, but it's the best we can do
         if PyCrypto is not available safely.
         """
+        #  The signedimp.tools.get_bootstrap_code() function will inline the
+        #  raw code from signedimp.cryptobase.rsa
         if not _signedimp_util._RSAKeyWithPSS:
             from signedimp.cryptobase.rsa import RSAKeyWithPSS
             _signedimp_util._RSAKeyWithPSS = RSAKeyWithPSS
@@ -224,58 +222,50 @@ class _signedimp_util:
         """Try to make the best versions of the utility funcs we can.
 
         This method should be called whenever a new set of imports become
-        available.  It will interrogate the runtime environment and try to
-        find better implementations of the utility functions defined on
-        this class.
+        available.  It will attempt to import and use better implementations
+        of the utility functions defined on this class.
         """
-        if _signedimp_mod_available("base64"):
-            try:
-                import base64
-                _signedimp_util.b64decode = staticmethod(base64.b64decode)
-            except ImportError:
-                pass
-        if _signedimp_mod_available("signedimp.crypto"):
-            # Awesome, we can use our crypto primitives directly
-            try:
-                import signedimp.crypto.md5
-                import signedimp.crypto.sha1
-                import signedimp.crypto.rsa
-                _signedimp_util.md5 = md5.md5
-                _signedimp_util.sha1 = sha1.sha1
-                _signedimp_util.RSAKeyWithPSS = rsa.RSAKeyWithPSS
-            except ImportError:
-                pass
-        elif _signedimp_mod_available("hashlib"):
-            # Great, we can use hashlib directly
+        #  Try to use native b64decode
+        try:
+            import base64
+            _signedimp_util.b64decode = staticmethod(base64.b64decode)
+        except ImportError:
+            pass
+        #  Try to use our fast-path crypto library
+        try:
+            from signedimp.crypto import md5
+            from signedimp.crypto import sha1
+            from signedimp.crypto import rsa
+            _signedimp_util.md5 = md5.md5
+            _signedimp_util.sha1 = sha1.sha1
+            _signedimp_util.RSAKeyWithPSS = rsa.RSAKeyWithPSS
+        except ImportError:
+            # Try to use hashlib
             try:
                 import hashlib
                 _signedimp_util.md5 = hashlib.md5
                 _signedimp_util.sha1 = hashlib.sha1
             except ImportError:
-                pass
-        elif _signedimp_mod_available("_hashlib"):
-            # Good, we can use the exposed openssl interface directly
-            try:
-                import _hashlib
-                _signedimp_util.md5 = _hashlib.openssl_md5
-                _signedimp_util.sha1 = _hashlib.openssl_sha1
-            except (ImportError,AttributeError):
-                pass
-        else:
-            if _signedimp_mod_available("_md5"):
-                #  OK, at least md5 is builtin
-                import _md5
-                _signedimp_util.md5 = _md5.new
-            if _signedimp_mod_available("_sha"):
-                #  OK, at least sha1 is builtin
-                import _sha
-                _signedimp_util.sha1 = _sha.new
+                # Try to use _hashlib
+                try:
+                    import _hashlib
+                    _signedimp_util.md5 = _hashlib.openssl_md5
+                    _signedimp_util.sha1 = _hashlib.openssl_sha1
+                except (ImportError,AttributeError):
+                    #  Try to use _md5 and _sha
+                    try:
+                        import _md5
+                        _signedimp_util.md5 = _md5.new
+                    except ImportError:
+                        pass
+                    try:
+                        import _sha
+                        _signedimp_util.sha1 = _sha.new
+                    except ImportError:
+                        pass
         #  If all else fails, we've left them as pure-python implementations
 
 
-# We might be lucky and have some modules availabe...
-_signedimp_util.recreate()
-
 
 class SignedHashDatabase(object):
     """An in-memory database of verified hash data.
@@ -386,19 +376,27 @@ class SignedImportManager(object):
     def __init__(self,valid_keys=[]):
         self.valid_keys = [k for k in valid_keys]
         self.hashdb = SignedHashDatabase(self.valid_keys)
+        self._hashdb_cache = {}
+
+    def add_valid_key(self,key):
+        self.valid_keys.append(key)
+        self.reinstall()
 
     def install(self):
         """Install this manager into the process import machinery."""
         if self not in sys.meta_path:
             sys.meta_path.insert(0,self)
+            self._orig_load_dynamic = imp.load_dynamic
+            self._orig_load_compiled = imp.load_compiled
+            self._orig_load_source = imp.load_source
+            imp.load_dynamic = self._imp_load_dynamic
+            imp.load_compiled = self._imp_load_compiled
+            imp.load_source = self._imp_load_source
+        self.reinstall()
+
+    def reinstall(self):
+        """Notify the manager that new imports may be available."""
         #  Try to speed things up by loading faster crypto primitives.
-        for mod in ("signedimp.crypto",):
-            try:
-                loader = self.find_module(mod)
-                if loader is not None:
-                    loader.load_module(mod)
-            except (ImportError,IntegrityCheckMissing,):
-                pass
         _signedimp_util.recreate()
 
     def find_module(self,fullname,path=None):
@@ -462,7 +460,86 @@ class SignedImportManager(object):
             if importer is None:
                 importer = DefaultImporter(path)
         return importer
-        
+
+    def _find_hashdb(self,path):
+        """Find and return the hashdb covering the given path."""
+        path = os.path.dirname(path)
+        while True:
+            hashfile = os.path.join(path,HASHFILE_NAME)
+            if hashfile in self._hashdb_cache:
+                return path,self._hashdb_cache[hashfile]
+            if os.path.exists(hashfile):
+                f = open(hashfile,"rb")
+                try:
+                    hashdata = f.read()
+                finally:
+                    f.close()
+                hashdb = SignedHashDatabase(self.valid_keys)
+                hashdb.parse_hash_data(hashdata)
+                self._hashdb_cache[hashfile] = hashdb
+                return path,hashdb
+            new_path = os.path.dirname(path)
+            if path == new_path:
+                break
+            path = new_path
+
+    def _imp_load_dynamic(self,name,pathname,file=None):
+        """Replacement for imp.load_dynamic."""
+        (basepath,hashdb) = self._find_hashdb(pathname)
+        if hashdb is None:
+            raise IntegrityCheckMissing(pathname)
+        f = open(pathname,"rb")
+        try:
+            data = f.read()
+        finally:
+            f.close()
+        try:
+            hashdb.verify("m",name,data)
+        except IntegrityCheckError:
+            hashdb.verify("d",pathname[len(basepath)+1:],data)
+        if file is not None:
+            return self._orig_load_dynamic(name,pathname,file)
+        else:
+            return self._orig_load_dynamic(name,pathname)
+
+    def _imp_load_compiled(self,name,pathname,file=None):
+        """Replacement for imp.load_compiled."""
+        (basepath,hashdb) = self._find_hashdb(pathname)
+        if hashdb is None:
+            raise IntegrityCheckMissing(pathname)
+        f = open(pathname,"rb")
+        try:
+            data = f.read()
+        finally:
+            f.close()
+        try:
+            hashdb.verify("m",name,data)
+        except IntegrityCheckError:
+            hashdb.verify("d",pathname[len(basepath)+1:],data)
+        if file is not None:
+            return self._orig_load_compiled(name,pathname,file)
+        else:
+            return self._orig_load_compiled(name,pathname)
+
+    def _imp_load_source(self,name,pathname,file=None):
+        """Replacement for imp.load_source."""
+        (basepath,hashdb) = self._find_hashdb(pathname)
+        if hashdb is None:
+            raise IntegrityCheckMissing(pathname)
+        f = open(pathname,"rb")
+        try:
+            data = f.read()
+        finally:
+            f.close()
+        try:
+            hashdb.verify("m",name,data)
+        except IntegrityCheckError:
+            hashdb.verify("d",pathname[len(basepath)+1:],data)
+        if file is not None:
+            return self._orig_load_source(name,pathname,file)
+        else:
+            return self._orig_load_source(name,pathname)
+
 
 class SignedLoader:
     """Wrapper class for managing signed imports from a specific loader.
@@ -476,13 +553,22 @@ class SignedLoader:
     def __init__(self,manager,loader):
         self.manager = manager
         self.loader = loader
-        self.hashdb = SignedHashDatabase(manager.valid_keys)
         try:
-            hashdata = self.loader.get_data(HASHFILE_NAME)
+            hashfile = self.get_datafilepath(HASHFILE_NAME)
+            if hashfile is None:
+                    hashdata = self.loader.get_data(HASHFILE_NAME)
+                    self.hashdb = SignedHashDatabase(manager.valid_keys)
+                    self.hashdb.parse_hash_data(hashdata)
+            else:
+                try:
+                    self.hashdb = manager._hashdb_cache[hashfile]
+                except KeyError:
+                    hashdata = self.loader.get_data(HASHFILE_NAME)
+                    self.hashdb = SignedHashDatabase(manager.valid_keys)
+                    self.hashdb.parse_hash_data(hashdata)
+                    manager._hashdb_cache[hashfile] = self.hashdb
         except (AttributeError,EnvironmentError):
-            pass
-        else:
-            self.hashdb.parse_hash_data(hashdata)
+            self.hashdb = SignedHashDatabase(manager.valid_keys)
 
     def __getattr__(self,attr):
         """Pass through simple attributes of the wrapped loader.
@@ -534,7 +620,15 @@ class SignedLoader:
 
     def is_package(self,fullname):
         """Check whether the given module is a package."""
-        return self.loader.is_package(fullname)
+        #  The py2exe ZipExtensionLoader (at least) seems to be broken here,
+        #  raising an ImportError for bundled DLLs.  We do a double-check then
+        #  just report False.
+        try:
+            return self.loader.is_package(fullname)
+        except ImportError:
+            if self.loader.find_module(fullname) is not None:
+                return False
+            raise
 
     def get_code(self,fullname):
         """Get the code object for the given module.
@@ -558,13 +652,32 @@ class SignedLoader:
         """Get the filename associated with the given module."""
         return self.loader.get_filename(fullname)
 
+    def get_datafilepath(self,path):
+        """Get the full path for the given data file.
+
+        If a full path cannot be constructed, None is returned.
+
+        This isn't an official PEP-302 method, so we hack it ourselves for the
+        standard import hooks and leave the option open to third-party modules
+        to implement it if they wish.
+        """
+        try:
+            gdfp = self.loader.get_datafilepath
+        except AttributeError:
+            try:
+                return os.path.join(self.loader.archive,path)
+            except AttributeError:
+                return None
+        else:
+            return gdfp(path)
+
     def verify_module(self,fullname,data):
         """Verify the given module data against our signature database."""
         self._verify("m",fullname,data)
 
     def verify_data(self,path,data):
         """Verify the given extra data against our signature database."""
-        self._verify("d",fullname,data)
+        self._verify("d",path,data)
 
     def _verify(self,typ,name,data):
         """Verify date for the given type specifier and name.
@@ -627,7 +740,7 @@ class BuiltinImporter(object):
     """
 
     @classmethod
-    def find_module(self,fullname):
+    def find_module(self,fullname,path=None):
         if imp.is_builtin(fullname):
             return self
         if imp.is_frozen(fullname):
@@ -854,10 +967,49 @@ class _DefaultImporter:
         """
         if self.base_path is None:
             raise OSError
-        return open(os.path.join(self.base_path,path),"rb").read()
+        f = open(os.path.join(self.base_path,path),"rb")
+        try:
+            return f.read()
+        finally:
+            f.close()
+
+    def get_filename(self,fullname):
+        """Get the code object for the given module."""
+        file,pathname,description = self._get_module_info(fullname)
+        if file is not None:
+            file.close()
+        return pathname
+
+    def get_datafilepath(self,path):
+        """Get the full path for the given data file."""
+        if self.base_path is None:
+            return None
+        return os.path.join(self.base_path,path)
 
 
 
 def RSAKeyWithPSS(*args,**kwds):
+    """Wrapper to expose RSAKeyWithPSS at the top-level of the module."""
     return _signedimp_util.RSAKeyWithPSS(*args,**kwds)
 
+
+#  Try to speed up initial imports by loading builtin or frozen utility mods.
+#  We forcibly block import of non-builtin-or-frozen modules, so that we
+#  can load frozen modules without fearing that they'll import something 
+#  unsafe under the hood.
+
+class BuiltinOnlyImporter(BuiltinImporter):
+    """Meta-path hook to only permit import of builtin or frozen modules.
+
+    This import hook claims to be able to import any module, but will succeed
+    only for builtin or frozen modules.
+    """
+    @classmethod
+    def find_module(self,fullname,path=None):
+        return self
+
+sys.meta_path.insert(0,BuiltinOnlyImporter)
+_signedimp_util.recreate()
+del sys.meta_path[0]
+
+
