@@ -92,7 +92,6 @@ class TestSignedImp_DefaultImport(unittest.TestCase):
                         "sim.install()",
                         "import signedimp_test.test1",
                         "print signedimp_test.test1.value")
-        print p.stderr.read()
         self.assertEquals(p.wait(),0)
         self.assertEquals(p.stdout.read().strip(),"7")
  
@@ -150,7 +149,6 @@ class TestSignedImp_DefaultImport(unittest.TestCase):
                         "sim.install()",
                         "import signedimp_test.test1",
                         "print signedimp_test.test1.value")
-        print p.stderr.read()
         self.assertEquals(p.wait(),0)
         self.assertEquals(p.stdout.read().strip(),"7")
         p = self._runpy("from signedimp import SignedImportManager",
@@ -266,6 +264,19 @@ else:
             p = popen(os.path.join(self.distdir,"script.exe"))
             self.assertEquals(p.wait(),0)
 
+        def test_replacement_load_dynamic_is_called(self):
+            #  py2exe loads .pyd via a stub that calls imp.load_dynamic().
+            #  Corrupt one and make sure it's detected as invalid.
+            signedimp.tools.sign_py2exe_app(self.distdir)
+            dynlib = "Crypto.PublicKey._fastmath.pyd"
+            with open(os.path.join(self.distdir,dynlib),"ab") as f:
+                f.write("malicious code")
+            p = popen(os.path.join(self.distdir,"script.exe"))
+            err = p.stderr.read()
+            self.assertNotEquals(p.wait(),0)
+            self.assertTrue("invalid hash" in err)
+            self.assertTrue("Crypto.PublicKey._fastmath" in err)
+
 try:
     import cx_Freeze
 except ImportError:
@@ -303,7 +314,6 @@ else:
         def test_signed_app_succeeds(self):
             signedimp.tools.sign_cxfreeze_app(self.distdir)
             p = popen(self.scriptexe)
-            print p.stderr.read()
             self.assertEquals(p.wait(),0)
 
         def test_unsigned_app_fails(self):
@@ -353,41 +363,42 @@ else:
             st_setup(name="testapp",version="0.1",app=[scriptfile],
                        options={"bdist":{"dist_dir":distdir}},
                        script_args=["py2app"])
+            self.appdir = os.path.join(self.distdir,"testapp.app")
 
         def tearDown(self):
             shutil.rmtree(self.tdir)
 
         def test_the_test(self):
-            p = popen(os.path.join(self.distdir,"testapp.app/Contents/MacOS/testapp"))
+            p = popen(os.path.join(self.appdir,"Contents/MacOS/testapp"))
             self.assertEquals(p.wait(),0)
 
         def test_signed_app_succeeds(self):
-            signedimp.tools.sign_py2exe_app(self.distdir)
-            p = popen(os.path.join(self.distdir,"testapp.app/Contents/MacOS/testapp"))
+            signedimp.tools.sign_py2app_bundle(self.appdir)
+            p = popen(os.path.join(self.appdir,"Contents/MacOS/testapp"))
             self.assertEquals(p.wait(),0)
 
         def test_unsigned_app_fails(self):
-            signedimp.tools.sign_py2exe_app(self.distdir)
-            zf = zipfile.ZipFile(os.path.join(self.distdir,"testapp.app/Contents/Resources/lib/python%d.%d/site-packages.zip" % sys.version_info[:2]),"a")
+            signedimp.tools.sign_py2app_bundle(self.appdir)
+            zf = zipfile.ZipFile(os.path.join(self.appdir,"Contents/Resources/lib/python%d.%d/site-packages.zip" % sys.version_info[:2]),"a")
             zf.writestr(signedimp.HASHFILE_NAME,"")
-            p = popen(os.path.join(self.distdir,"testapp.app/Contents/MacOS/testapp"))
+            p = popen(os.path.join(self.appdir,"Contents/MacOS/testapp"))
             self.assertNotEquals(p.wait(),0)
 
         def test_modified_app_fails(self):
-            signedimp.tools.sign_py2exe_app(self.distdir)
-            zf = zipfile.ZipFile(os.path.join(self.distdir,"testapp.app/Contents/Resources/lib/python%d.%d/site-packages.zip" % sys.version_info[:2]),"a")
+            signedimp.tools.sign_py2app_bundle(self.appdir)
+            zf = zipfile.ZipFile(os.path.join(self.appdir,"Contents/Resources/lib/python%d.%d/site-packages.zip" % sys.version_info[:2]),"a")
             zf.writestr("signedimp/crypto/__init__.py","")
-            p = popen(os.path.join(self.distdir,"testapp.app/Contents/MacOS/testapp"))
+            p = popen(os.path.join(self.appdir,"Contents/MacOS/testapp"))
             self.assertNotEquals(p.wait(),0)
 
         def test_unchecked_modules_fails(self):
-            signedimp.tools.sign_py2exe_app(self.distdir,check_modules=[])
-            p = popen(os.path.join(self.distdir,"testapp.app/Contents/MacOS/testapp"))
+            signedimp.tools.sign_py2app_bundle(self.appdir,check_modules=[])
+            p = popen(os.path.join(self.appdir,"Contents/MacOS/testapp"))
             self.assertNotEquals(p.wait(),0)
 
         def test_disabled_check_modules_succeeds(self):
-            signedimp.tools.sign_py2exe_app(self.distdir,check_modules=False)
-            p = popen(os.path.join(self.distdir,"testapp.app/Contents/MacOS/testapp"))
+            signedimp.tools.sign_py2app_bundle(self.appdir,check_modules=False)
+            p = popen(os.path.join(self.appdir,"Contents/MacOS/testapp"))
             self.assertEquals(p.wait(),0)
 
 class TestMisc(unittest.TestCase):
