@@ -69,7 +69,7 @@ def get_bootstrap_code(indent=""):
         mod = __import__(mod,fromlist=["*"])
         src = inspect.getsource(mod)
         for ln in src.split("\n"):
-            if "from signedimp.cryptobase." in ln:
+            if ln.strip().startswith("from signedimp.cryptobase."):
                 lnstart = ln.find("from")
                 newindent = indent + ln[:lnstart]
                 newmod = ln.strip()[5:].split()[0]
@@ -223,37 +223,68 @@ def sign_py2exe_app(appdir,key=None,hash="sha1",check_modules=None):
     """
     if check_modules is None:
         check_modules = ["_memimporter"]
+    do_check_modules = (check_modules != False)
     #  Since the public key will be embedded in the executables, it's OK to
     #  generate a throw-away key that's purely for signing this particular app.
     if key is None:
         key = RSAKeyWithPSS.generate()
+    pubkey = key.get_public_key()
     #  Built the bootstrapping code needed for each executable.
     #  We init the bootstrap objects inside a function so they get their own
     #  namespace; py2exe's own bootstrap code does a "del sys" which would
     #  play havoc with the import machinery.
+    bscodestr = get_bootstrap_code(indent="        ")
     bscode =  """
 import sys
-def _signedimp_init():
-    %s
-    class _signedimp_exports:
-        pass
-    for nm in __all__:
-        setattr(_signedimp_exports,nm,staticmethod(locals()[nm]))
-    return _signedimp_exports
-signedimp = _signedimp_init()
-if %s:
+
+#  Try to figure out if signed imports are already enabled.
+#  We can't use isinstance() since there may be multiple copies of the
+#  signedimp module created by different bootstrapping codes.
+try:
+    if sys.meta_path[0].__class__.__name__ == "SignedImportManager":
+        signedimp_installed = True
+    else:
+        signedimp_installed = False
+except (IndexError,AttributeError):
+    signedimp_installed = False
+
+#  Check the boot-time modules if necessary.
+if %(do_check_modules)r and not signedimp_installed:
     for mod in sys.modules:
-        if mod not in sys.builtin_module_names and mod not in %s:
+        if mod in sys.builtin_module_names:
+            continue
+        if mod not in %(check_modules)r:
             err = "module '%%s' already loaded, integrity checks impossible"
             sys.stderr.write(err %% (mod,))
             sys.stderr.write("\\nTerminating the program.\\n")
             sys.exit(1)
-k = signedimp.%s
-signedimp.SignedImportManager([k]).install()
-""" % (get_bootstrap_code(indent="    "),
-       check_modules != False,
-       repr(check_modules),
-       repr(key.get_public_key()),)
+
+#  Get a reference to the signedimp module, possibly by creating
+#  a private copy from raw code.
+signedimp = None
+if signedimp_installed:
+    try:
+        import signedimp
+    except ImportError:
+        pass
+if signedimp is None:
+    def _signedimp_init():
+        %(bscodestr)s
+        class _signedimp_exports:
+            pass
+        for nm in __all__:
+            setattr(_signedimp_exports,nm,staticmethod(locals()[nm]))
+        return _signedimp_exports
+    signedimp = _signedimp_init()
+
+#  Add the specific key into the signed import machinery.
+k = signedimp.%(pubkey)r
+if signedimp_installed:
+    sys.meta_path[0].add_valid_key(k)
+else:
+    signedimp.SignedImportManager([k]).install()
+
+""" % locals()
     bscode = compile(bscode,"__main__.py","exec")
     #  Hack the bootstrap code into the start of each script to be run.
     #  This unfortunately depends on some inner details of the py2exe format.
@@ -310,36 +341,70 @@ def sign_py2app_bundle(appdir,key=None,hash="sha1",check_modules=None):
     """
     if check_modules is None:
         check_modules = ["codecs","encodings","encodings.__builtin__",
-                         "encodings.codecs","encodings.utf_8",
-                         "encodings.aliases","encodings.encodings","readline"]
+                         "encodings.codecs","encodings.utf_8","copy_reg","site",
+                         "abc","os","wxhack","posixpath","_abcoll","os.path",
+                         "genericpath","stat","warnings","types","linecache",
+                         "encodings.aliases","encodings.encodings","readline",
+                         "UserDict","zlib"]
+    do_check_modules = (check_modules != False)
     #  Since the public key will be embedded in the executables, it's OK to
     #  generate a throw-away key that's purely for signing this particular app.
     if key is None:
         key = RSAKeyWithPSS.generate()
+    pubkey = key.get_public_key()
     #  Build the bootstrap code and put it at start of __boot__.py.
+    bscodestr = get_bootstrap_code(indent="        ")
     bscode =  """
 import sys
-def _signedimp_init():
-    %s
-    class _signedimp_exports:
-        pass
-    for nm in __all__:
-        setattr(_signedimp_exports,nm,staticmethod(locals()[nm]))
-    return _signedimp_exports
-signedimp = _signedimp_init()
-if %s:
+
+#  Try to figure out if signed imports are already enabled.
+#  We can't use isinstance() since there may be multiple copies of the
+#  signedimp module created by different bootstrapping codes.
+try:
+    if sys.meta_path[0].__class__.__name__ == "SignedImportManager":
+        signedimp_installed = True
+    else:
+        signedimp_installed = False
+except (IndexError,AttributeError):
+    signedimp_installed = False
+
+#  Check the boot-time modules if necessary.
+if %(do_check_modules)r and not signedimp_installed:
     for mod in sys.modules:
-        if mod not in sys.builtin_module_names and mod not in %s:
+        if mod in sys.builtin_module_names:
+            continue
+        if mod not in %(check_modules)r:
             err = "module '%%s' already loaded, integrity checks impossible"
             sys.stderr.write(err %% (mod,))
             sys.stderr.write("\\nTerminating the program.\\n")
             sys.exit(1)
-k = signedimp.%s
-signedimp.SignedImportManager([k]).install()
-""" % (get_bootstrap_code(indent="    "),
-       check_modules != False,
-       repr(check_modules),
-       repr(key.get_public_key()),)
+
+#  Get a reference to the signedimp module, possibly by creating
+#  a private copy from raw code.
+signedimp = None
+if signedimp_installed:
+    try:
+        import signedimp
+    except ImportError:
+        pass
+if signedimp is None:
+    def _signedimp_init():
+        %(bscodestr)s
+        class _signedimp_exports:
+            pass
+        for nm in __all__:
+            setattr(_signedimp_exports,nm,staticmethod(locals()[nm]))
+        return _signedimp_exports
+    signedimp = _signedimp_init()
+
+#  Add the specific key into the signed import machinery.
+k = signedimp.%(pubkey)r
+if signedimp_installed:
+    sys.meta_path[0].add_valid_key(k)
+else:
+    signedimp.SignedImportManager([k]).install()
+
+""" % locals()
     bsfile = os.path.join(appdir,"Contents","Resources","__boot__.py")
     with open(bsfile,"r+") as f:
         oldcode = f.read()
@@ -349,8 +414,9 @@ signedimp.SignedImportManager([k]).install()
     #  Sign the main library.zip
     libdir = os.path.join(appdir,"Contents","Resources","lib")
     libdir = os.path.join(libdir,"python%d.%d"%sys.version_info[:2])
-    libzip = os.path.join(libdir,"site-packages.zip")
-    sign_zipfile(libzip,key,hash=hash)
+    for nm in os.listdir(libdir):
+        if nm.endswith(".zip"):
+            sign_zipfile(os.path.join(libdir,nm),key,hash=hash)
     #  Sign a variety of potential code dirs
     try:
         sign_directory(os.path.join(libdir,"lib-dynload"),key,hash=hash)
@@ -398,34 +464,59 @@ def sign_cxfreeze_app(appdir,key=None,hash="sha1",check_modules=None):
         check_modules = ["codecs","encodings","encodings.__builtin__",
                          "encodings.codecs","encodings.utf_8",
                          "encodings.aliases","encodings.encodings"]
+    do_check_modules = (check_modules != False)
     #  Since the public key will be embedded in the executables, it's OK to
     #  generate a throw-away key that's purely for signing this particular app.
     if key is None:
         key = RSAKeyWithPSS.generate()
+    pubkey = key.get_public_key()
     #  Build the bootstrap code to be inserted into each executable.  Since
     #  it replaces the cx_Freeze__init__ script it needs to exec that once
     #  the signed imports are in place.
     bscode_tmplt =  """
 import sys
-if %s:
+
+#  Try to figure out if signed imports are already enabled.
+#  We can't use isinstance() since there may be multiple copies of the
+#  signedimp module created by different bootstrapping codes.
+try:
+    if sys.meta_path[0].__class__.__name__ == "SignedImportManager":
+        signedimp_installed = True
+    else:
+        signedimp_installed = False
+except (IndexError,AttributeError):
+    signedimp_installed = False
+
+#  Check the boot-time modules if necessary.
+if %(do_check_modules)r and not signedimp_installed:
     for mod in sys.modules:
-        if mod not in sys.builtin_module_names and mod not in %s:
+        if mod == "signedimp" or mod.startswith("signedimp."):
+            continue
+        if mod in sys.builtin_module_names:
+            continue
+        if mod not in %(check_modules)r:
             err = "module '%%s' already loaded, integrity checks impossible"
             sys.stderr.write(err %% (mod,))
             sys.stderr.write("\\nTerminating the program.\\n")
             sys.exit(1)
 
+#  Since it's bundled into a zipfile, we always get at signedimp by
+#  just importing it.  This will re-use an existing version if loaded.
 import signedimp
-k = signedimp.%s
-signedimp.SignedImportManager([k]).install()
 
-print FILE_NAME
-print SHARED_ZIP_FILE_NAME
-print EXCLUSIVE_ZIP_FILE_NAME
-print INITSCRIPT_ZIP_FILE_NAME
-if %s:
+#  Add the specific key into the signed import machinery.
+k = signedimp.%(pubkey)r
+if signedimp_installed:
+    sys.meta_path[0].add_valid_key(k)
+else:
+    signedimp.SignedImportManager([k]).install()
+
+#  Bootstrap the original cx_Freeze__init__ module.
+#  If it was in the appended zipfile, we're given it as a marshalled string.
+#  If not, we need to search for it in the other zipfiles.
+if %(has_initcode)r:
     import marshal
-    exec marshal.loads(%s)
+    exec marshal.loads(%(initcode)r)
 else:
     import zipimport
     initmod = "cx_Freeze__init__"
@@ -451,16 +542,12 @@ else:
             #  If it contains the init module already, we'll need to
             #  grab its code to bootstrap into it.
             try:
-                initcode = zf.read(initmod+".pyc")[8:]
+                initcode = repr(zf.read(initmod+".pyc")[8:])
             except KeyError:
                 initcode = ""
             #  Store our own code as the cxfreeze init module
-            bssrc = bscode_tmplt % (check_modules != False,
-                                    repr(check_modules),
-                                    repr(key.get_public_key()),
-                                    bool(initcode),
-                                    repr(initcode),
-                    )
+            has_initcode = bool(initcode)
+            bssrc = bscode_tmplt % locals()
             bscode = imp.get_magic() + struct.pack("<i",time.time())
             bscode += marshal.dumps(compile(bssrc,initmod+".py","exec"))
             zf.writestr(initmod+".pyc",bscode)
